# frozen_string_literal: true

require 'ractor/local_variable'
require_relative 'channel/version'

class Ractor
  module Channel
    SAVED_BY_CHANNEL = Ractor::LocalVariable.new(delegate: true) do
      mutex = ::Mutex.new
      Hash.new do |h, channel|
        mutex.synchronize do
          h.fetch(channel) { h[channel] = ::Queue.new }
        end
      end
    end

    def send(value, channel: nil, **options)
      super([channel, value].freeze, **options)
    end

    module ClassMethods
      def receive(channel: nil) # rubocop:disable Metrics/MethodLength (https://bugs.ruby-lang.org/issues/17374)
        if channel == :*
          SAVED_BY_CHANNEL.each do |_channel, queue|
            return queue.pop(true)
          rescue ThreadError
            # try next channel
          end
          super() # # FIXME: Not thread-safe
        else
          loop do
            return SAVED_BY_CHANNEL[channel].pop(true)
          rescue ThreadError
            dest_channel, message = super()
            SAVED_BY_CHANNEL[dest_channel] << message
          end
        end
      end
    end

    def self.prepended(base)
      base.singleton_class.prepend(ClassMethods)
    end

    refine Ractor do
      prepend Ractor::Channel
    end

    refine Ractor.singleton_class do
      prepend Ractor::Channel::ClassMethods
    end
  end
end
